from sourpea.primitives import DerivedLevel, Level, Factor, Block, WithinTrialDerivationWindow, \
    TransitionDerivationWindow

# valid test sequence
test_sequence_1 = [
    {'word': 'red', 'color': 'red', 'congruency': 'con'},
    {'word': 'red', 'color': 'green', 'congruency': 'inc'},
    {'word': 'green', 'color': 'red', 'congruency': 'inc'},
    {'word': 'green', 'color': 'green', 'congruency': 'con'},
]

# invalid test sequence: derived level mixup
test_sequence_2 = [
    {'word': 'red', 'color': 'red', 'congruency': 'inc'},
    {'word': 'red', 'color': 'green', 'congruency': 'con'},
    {'word': 'green', 'color': 'red', 'congruency': 'con'},
    {'word': 'green', 'color': 'green', 'congruency': 'inc'},
]

# invalid test sequence: not counterbalanced
test_sequence_3 = [
    {'word': 'red', 'color': 'green', 'congruency': 'inc'},
    {'word': 'red', 'color': 'green', 'congruency': 'inc'},
    {'word': 'green', 'color': 'green', 'congruency': 'con'},
    {'word': 'green', 'color': 'green', 'congruency': 'con'},
]

word = Factor('word', ['red', 'green'])
color = Factor('color', ['red', 'green'])


def is_con(w, c):
    return w == c


def is_inc(w, c):
    return not is_con(w, c)


congruent = DerivedLevel('con', window=WithinTrialDerivationWindow(predicate=is_con, factors=[word, color]))
incongruent = DerivedLevel('inc', window=WithinTrialDerivationWindow(predicate=is_inc, factors=[word, color]))

congruency = Factor('congruency', [congruent, incongruent])

congruency_2 = Factor('congruency', ['con', 'inc'])

block = Block(design=[word, color, congruency], crossing=[word, congruency])
test_1 = block.test(test_sequence_1)
test_2 = block.test(test_sequence_2)
test_3 = block.test(test_sequence_3)

print(test_1['pValue'], test_1['levels'])
print(test_2['pValue'], test_2['levels'])
print(test_3['pValue'], test_3['levels'])

test_sequence_4 = [
    {'task': 'color naming', 'congruency': 'congruent', 'transition': None},
    {'task': 'color naming', 'congruency': 'congruent', 'transition': 'full repetition'},
    {'task': 'word naming', 'congruency': 'congruent', 'transition': 'half repetition'},
    {'task': 'color naming', 'congruency': 'incongruent', 'transition': 'switch'}
]

test_sequence_5 = [
    {'task': 'color naming', 'congruency': 'congruent', 'transition': None},
    {'task': 'color naming', 'congruency': 'congruent', 'transition': 'full repetition'},
    {'task': 'color naming', 'congruency': 'incongruent', 'transition': 'half repetition'},
    {'task': 'word naming', 'congruency': 'congruent', 'transition': 'switch'},
    {'task': 'word naming', 'congruency': 'congruent', 'transition': 'full repetition'},
    {'task': 'word naming', 'congruency': 'incongruent', 'transition': 'half repetition'},
    {'task': 'color naming', 'congruency': 'congruent', 'transition': 'switch'}
]

test_sequence_6 = [
    {'task': 'color naming', 'congruency': 'congruent', 'transition': None},
    {'task': 'color naming', 'congruency': 'congruent', 'transition': 'full repetition'},
    {'task': 'color naming', 'congruency': 'incongruent', 'transition': 'half repetition'},
    {'task': 'word naming', 'congruency': 'congruent', 'transition': 'switch'},
    {'task': 'word naming', 'congruency': 'congruent', 'transition': 'full repetition'},
    {'task': 'word naming', 'congruency': 'incongruent', 'transition': 'half repetition'},
    {'task': 'color naming', 'congruency': 'incongruent', 'transition': 'half repetition'}
]

task = Factor('task', ['color naming', 'word naming'])
congruency = Factor('congruency', ['congruent', 'incongruent'])


def is_full_repetition(t, c):
    return t[0] == t[1] and c[0] == c[1]


def is_switch(t, c):
    return t[0] != t[1] and c[0] != c[1]


def is_half_repetition(t, c):
    return not (is_full_repetition(t, c) or is_switch(t, c))


full_repetition = DerivedLevel('full repetition',
                               TransitionDerivationWindow(predicate=is_full_repetition, factors=[task, congruency]))
switch = DerivedLevel('switch',
                      TransitionDerivationWindow(predicate=is_switch, factors=[task, congruency]))
half_repetition = DerivedLevel('half repetition',
                               TransitionDerivationWindow(predicate=is_half_repetition, factors=[task, congruency]))

transition = Factor('transition', [full_repetition, switch, half_repetition])

block_4 = Block(design=[task, congruency, transition], crossing=[transition])
test_4 = block_4.test(test_sequence_4)

print(test_4['pValue'], test_4['levels'])

block_5 = Block(design=[task, congruency, transition], crossing=[task, transition])
test_5 = block_5.test(test_sequence_5)

print(test_5['pValue'], test_5['levels'])

block_6 = Block(design=[task, congruency, transition], crossing=[task, transition])
test_6 = block_6.test(test_sequence_6)

print(test_6['pValue'], test_6['levels'])



import pandas as pd


def trials_from_csv(path):
    """
    :param path: the path to a csv file in wide format
    :return: a list were every item is a full trial
    """
    df = pd.read_csv(path)
    dictionary = df.to_dict('list')
    return _long(dictionary)


def trials_from_sweetpea(sequence):
    """
    :param sequence: a trial sequence generated by sweetpea (one experiment out of experiments!)
    :return: a list were every item is a full trial
    """
    return _long(sequence)


def _long(sequence):
    """
    :param sequence: a trial list dictionary in wide format
    :return: a list were every item is a full trial
    """
    keys = []
    for key in sequence.keys():
        keys.append(key)
    long = []
    for i in range(len(sequence[keys[0]])):
        t = {}
        for key in keys:
            t[key] = sequence[key][i]
        long.append(t)
    return long





from sweetpea import Factor, DerivedLevel, Window, MultiCrossBlock, synthesize_trials, Exclude, MinimumTrials, \
    CrossBlock, save_experiments_csv


def create_two_back():
    letter = Factor('letter', ['b', 'f', 'm', 'q', 'k', 'x', 'r', 'h'])

    def is_target(letters):
        if not letters:
            return False
        return letters[-2] == letters[0]

    def is_not_target(letters):
        if not letters:
            return True
        return not is_target(letters)

    one_t = DerivedLevel('1', Window(is_target, [letter], width=3, stride=1, start=0), 1)
    two_t = DerivedLevel('2', Window(is_not_target, [letter], width=3, stride=1, start=0), 5)

    target = Factor('target', [one_t, two_t])

    # def is_one_back(letters):
    #     if letters is None:
    #         return False
    #     return letters[-1] == letters[0]
    #
    # def is_not_one_back(letters):
    #     return not is_one_back(letters)
    #
    # two_o = DerivedLevel('2', Window(is_one_back, [letter], width=2, stride=1, start=1), 1)
    # one_o = DerivedLevel('1', Window(is_not_one_back, [letter], width=2, stride=1, start=1), 5)
    #
    # one_back = Factor('one_back', [one_o, two_o])
    #
    # def is_three_back(letters):
    #     if letters is None:
    #         return False
    #     return letters[-3] == letters[0]
    #
    # def is_not_three_back(letters):
    #     return not is_three_back(letters)
    #
    # one_three = DerivedLevel('1', Window(is_three_back, [letter], width=4, stride=1, start=1))
    # zero_three = DerivedLevel('0', Window(is_not_three_back, [letter], width=4, stride=1, start=1))
    #
    # three_back = Factor('three_back', [one_three, zero_three])
    #
    # def is_control_target(letters):
    #     if letters is None:
    #         return False
    #     return is_target(letters) and letters[0] != letters[-1]
    #
    # def is_experimental_target(letters):
    #     if letters is None:
    #         return False;
    #     return is_target(letters) and letters[0] == letters[-1]
    #
    # def is_control_foil(letters):
    #     if letters is None:
    #         return False
    #     return is_not_target(letters) and letters[0] != letters[-1]
    #
    # def is_experimental_foil(letters):
    #     if letters is None:
    #         return True
    #     return is_not_target(letters) and letters[0] == letters[-1]
    #
    # one_one_0 = DerivedLevel('1/1/0', Window(is_control_target, [letter], width=3, stride=1, start=1), 3)
    # one_two_0 = DerivedLevel('1/2/0', Window(is_experimental_target, [letter], width=3, stride=1, start=1), 1)
    # two_one_0 = DerivedLevel('2/1/0', Window(is_control_foil, [letter], width=3, stride=1, start=1), 17)
    # two_two_0 = DerivedLevel('2/2/0', Window(is_experimental_foil, [letter], width=3, stride=1, start=1), 3)
    #
    # condi = Factor('condi', [one_one_0, one_two_0, two_one_0, two_two_0])

    block = CrossBlock(design=[letter, target],
                            constraints=[MinimumTrials(48)],
                            crossing=[letter, target])

    experiment = synthesize_trials(block, 1)

    save_experiments_csv(block, experiment, "sweetKaneEtAl/test")


create_two_back()



from sourpea.primitives import Factor, DerivedLevel, DerivationWindow, Block, TransitionDerivationWindow, Level, Exclude
import pandas as pd

PATHS = [f'kaneEtAl/ckm_2_back_{l}.txt' for l in ['a', 'b', 'c', 'd', 'e', 'f', 'g', 'h']]
PATHS_THREE = [f'kaneEtAl/ckm_3_back_{l}.txt' for l in ['a', 'b', 'c', 'd', 'e', 'f', 'g', 'h']]



def test_sequence_two_back(trial_sequence, name):
    letter = Factor('letter', ['b', 'f', 'm', 'q', 'k', 'x', 'r', 'h'])

    def is_target(letters):
        return letters[0] == letters[2]

    def is_not_target(letters):
        return not is_target(letters)

    one_t = DerivedLevel('1', DerivationWindow(is_target, [letter], 3), 8 / 48)
    two_t = DerivedLevel('2', DerivationWindow(is_not_target, [letter], 3), 40 / 48)

    target = Factor('target', [one_t, two_t])

    def is_one_back(letters):
        return letters[0] == letters[1]

    def is_not_one_back(letters):
        return not is_one_back(letters)

    two_o = DerivedLevel('2', DerivationWindow(is_one_back, [letter], 2), 8 / 48)
    one_o = DerivedLevel('1', DerivationWindow(is_not_one_back, [letter], 2), 40 / 48)

    one_back = Factor('one_back', [one_o, two_o])

    def is_three_back(letters):
        return letters[0] == letters[3]

    def is_not_three_back(letters):
        return not is_three_back(letters)

    one_three = DerivedLevel('1', DerivationWindow(is_three_back, [letter], 4), 1 / 48)
    zero_three = DerivedLevel('0', DerivationWindow(is_not_three_back, [letter], 4), 1)

    three_back = Factor('three_back', [one_three, zero_three])

    def is_control_target(letters):
        return is_target(letters) and letters[2] != letters[1]

    def is_experimental_target(letters):
        return is_target(letters) and letters[2] == letters[1]

    def is_control_foil(letters):
        return is_not_target(letters) and letters[2] != letters[1]

    def is_experimental_foil(letters):
        return is_not_target(letters) and letters[2] == letters[1]

    one_one_0 = DerivedLevel('1/1/0', DerivationWindow(is_control_target, [letter], 3), 6 / 48)
    one_two_0 = DerivedLevel('1/2/0', DerivationWindow(is_experimental_target, [letter], 3), 2 / 48)
    two_one_0 = DerivedLevel('2/1/0', DerivationWindow(is_control_foil, [letter], 3), 34 / 48)
    two_two_0 = DerivedLevel('2/2/0', DerivationWindow(is_experimental_foil, [letter], 3), 6 / 48)

    condi = Factor('condi', [one_one_0, one_two_0, two_one_0, two_two_0])

    block = Block(design=[letter, target, one_back], crossing=[letter, target])
    test = block.test(trial_sequence)

    block_2 = Block(design=[letter, one_back], crossing=[letter, one_back])
    test_2 = block_2.test(trial_sequence)

    block_3 = Block(design=[letter, condi], crossing=[letter, condi])
    test_3 = block_3.test(trial_sequence)

    block_4 = Block(design=[letter, three_back], constraints=[Exclude(three_back, '1')], crossing=[letter])
    test_4 = block_4.test(trial_sequence)

    block_5 = Block(design=[letter, condi], crossing=[condi])
    test_5 = block_5.test(trial_sequence)


    print(name)
    print(test)
    print(test_2)
    print(test_3)
    print(test_4)
    print(test_5)
    print('\n')


def sequence_two_txt(path):
    f = open(path)
    trials = []
    for line in f:
        if line[-8] != '\\':
            trials.append(
                {'letter': line[-8],
                 'target': line[-6].replace('\n', '').replace('\\', '/'),
                 'one_back': line[-4].replace('\n', '').replace('\\', '/'),
                 'three_back': line[-2].replace('\n', '').replace('\\', '/'),
                 'condi': line[-6:].replace('\n', '').replace('\\', '/')})
        else:
            trials.append({
                'letter': line[-7],
                'target': line[-5].replace('\n', '').replace('\\', '/'),
                'one_back': line[-3].replace('\n', '').replace('\\', '/'),
                'three_back': line[-1].replace('\n', '').replace('\\', '/'),
                'condi': line[-5:].replace('\n', '').replace('\\', '/')})
    return trials


def sequence_two_csv(path):
    df = pd.read_csv(path)
    letters = list(df['letter'])
    condi = list(df['condi'])
    sequence = []
    for l, c in zip(letters, condi):
        c = str(c)
        if not c == 'nan':
            sequence.append({'letter': l, 'target': c[0], 'one_back': c[2], 'three_back': c[-1], 'condi': c})
    return sequence


def sequence_three(path):
    f = open(path)
    trials = []
    for line in f:
        if line[-8] != '\\':
            trials.append(
                {'letter': line[-8],
                 'target': line[-6].replace('\n', '').replace('\\', '/'),
                 'one_or_two': line[-4].replace('\n', '').replace('\\', '/'),
                 'lure': line[-2].replace('\n', '').replace('\\', '/'),
                 'condi': line[-6:].replace('\n', '').replace('\\', '/')})
        else:
            trials.append({
                'letter': line[-7],
                'target': line[-5].replace('\n', '').replace('\\', '/'),
                'one_or_two': line[-3].replace('\n', '').replace('\\', '/'),
                'lure': line[-1].replace('\n', '').replace('\\', '/'),
                'condi': line[-5:].replace('\n', '').replace('\\', '/')})
    return trials


def test_three_back(PATH):
    trial_sequence = sequence_three(PATH)

    letter = Factor('letter', ['b', 'f', 'm', 'q', 'k', 'x', 'r', 'h'])

    def is_target(letters):
        return letters[0] == letters[3]

    def is_not_target(letters):
        return not is_target(letters)

    one_t = DerivedLevel('1', DerivationWindow(is_target, [letter], 4), 8 / 48)
    two_t = DerivedLevel('2', DerivationWindow(is_not_target, [letter], 4), 40 / 48)

    target = Factor('target', [one_t, two_t])

    def is_one_or_two(letters):
        return letters[1] == letters[2] or letters[0] == letters[2]

    def is_not_one_or_two(letters):
        return not is_one_or_two(letters)

    two_o = DerivedLevel('2', DerivationWindow(is_one_or_two, [letter], 3), 9 / 48)
    one_o = DerivedLevel('1', DerivationWindow(is_not_one_or_two, [letter], 3), 39 / 48)

    one_or_two = Factor('one_or_two', [one_o, two_o])

    def is_lure_one(letters):
        return not letters[0] == letters[3] and letters[3] == letters[2]

    def is_lure_two(letters):
        return not letters[0] == letters[3] and letters[3] == letters[1]

    def is_not_lure(letters):
        return not (is_lure_one(letters) or is_lure_two(letters))

    one_lure = DerivedLevel('1', DerivationWindow(is_lure_one, [letter], 4), 1 / 48)
    two_lure = DerivedLevel('2', DerivationWindow(is_lure_two, [letter], 4), 6 / 48)
    zero_lure = DerivedLevel('0', DerivationWindow(is_not_lure, [letter], 4), 41 / 48)

    lure = Factor('lure', [one_lure, two_lure, zero_lure])

    def is_control_target(letters):
        return letters[0] == letters[3] and letters[2] != letters[3] and letters[1] != letters[3]

    def is_experimental_target(letters):
        return letters[0] == letters[3] and letters[2] == letters[3]

    def is_control_foil(letters):
        return letters[0] != letters[3] and letters[1] != letters[3] and letters[2] != letters[3]

    def is_experimental_two_foil(letters):
        return letters[0] != letters[3] and letters[1] == letters[3]

    def is_experimental_one_foil(letters):
        return letters[0] != letters[3] and letters[2] == letters[3]

    one_one_0 = DerivedLevel('1/1/0', DerivationWindow(is_control_target, [letter], 4), 6 / 48)
    one_two_0 = DerivedLevel('1/2/0', DerivationWindow(is_experimental_target, [letter], 4), 2 / 48)
    two_one_0 = DerivedLevel('2/1/0', DerivationWindow(is_control_foil, [letter], 4), 33 / 48)
    two_two_two = DerivedLevel('2/2/2', DerivationWindow(is_experimental_two_foil, [letter], 4), 6 / 48)
    two_two_one = DerivedLevel('2/2/1', DerivationWindow(is_experimental_one_foil, [letter], 4), 1 / 48)

    condi = Factor('condi', [one_one_0, one_two_0, two_one_0, two_two_two, two_two_one])

    block = Block(design=[letter, target, one_or_two], crossing=[letter, target])
    test = block.test(trial_sequence)

    block_2 = Block(design=[letter, one_or_two], crossing=[one_or_two])
    test_2 = block_2.test(trial_sequence)

    block_3 = Block(design=[letter, lure], crossing=[lure])
    test_3 = block_3.test(trial_sequence)

    block_4 = Block(design=[letter, condi], crossing=[condi])
    test_4 = block_4.test(trial_sequence)

    print(PATH)
    print(test)
    print(test_2)
    print(test_3)
    print(test_4)
    print('\n')


def test_two_back_real(PATH):
    trial_sequence = sequence_two_txt(PATH)
    test_sequence_two_back(trial_sequence, PATH)


def test_two_back_csv(PATH):
    trial_sequence = sequence_two_csv(PATH)
    test_sequence_two_back(trial_sequence, PATH)


for p in PATHS:
    test_two_back_real(p)

# for p in PATHS_THREE:
#    test_three_back(p)

paths = [f'sweetKaneEtAl/test_{str(i)}_0.csv' for i in range(1,7)]
for p in paths:
    print(p)
    test_two_back_csv(p)



from __future__ import annotations
from typing import Callable, List, Tuple
from scipy.stats import chisquare


class Level:
    """
    A discrete value that a :class:`.Factor` can hold.
    """
    name: str
    weight: float = 1

    def __init__(self, name: str, weight: float = 1):
        self.name = name
        self.weight = weight


class DerivedLevel(Level):
    window: DerivationWindow

    def __init__(self, name: str, window: DerivationWindow, weight: float = 1):
        super().__init__(name, weight)
        self.window = window


class Factor:
    """"An independent variable in a factorial experiment. Factors are composed
    of :class:`Levels <.Level>` and come in two flavors:
    """
    name: str
    levels: List[Level]

    def __init__(self, name: str, initial_levels: list):
        self.name = name
        self.levels = []
        for level in initial_levels:
            if type(level) == str:
                self.levels.append(Level(level))
            elif isinstance(level, Level):
                self.levels.append(level)


class DerivationWindow:
    predicate: Callable
    factors: List[Factor]
    width: int

    def __init__(self, predicate: Callable, factors: List[Factor], width: int = 1):
        self.predicate = predicate
        self.factors = factors
        self.width = width


class WithinTrialDerivationWindow(DerivationWindow):

    def __init__(self, predicate: Callable, factors: List[Factor]):
        super().__init__(predicate, factors)
        self.width = 1


# Aliases
WithinTrial = WithinTrialDerivationWindow



class TransitionDerivationWindow(DerivationWindow):

    def __init__(self, predicate: Callable, factors: List[Factor]):
        super().__init__(predicate, factors)
        self.width = 2


# Aliases
Transition = TransitionDerivationWindow


class Constraint:

    def __init__(self):
        pass

    def test(self, sequence: List):
        return True


class Exclude(Constraint):
    factor: str
    level: str

    def __init__(self, factor: Factor, level: Level):
        if isinstance(factor, Factor):
            self.factor = factor.name
        else:
            self.factor = factor
        if isinstance(level, Level):
            self.level = level.name
        else:
            self.level = level

    def test(self, sequence: List):
        for trial in sequence:
            if trial[self.factor] == self.level:
                return False
        return True


class _NumberConstraint:
    trials: int

    def __init__(self, trials: int):
        super().__init__()
        self.trials = trials


class MinimumTrials(_NumberConstraint):

    def test(self, sequence: List):
        return len(sequence) >= self.trials


class _KConstraint(_NumberConstraint):
    factor: Factor
    levels: List[Level]

    def __init__(self, trials: int, level: Tuple[Factor, List[Level]]):
        super().__init__(trials)
        if isinstance(level, Factor):
            self.factor = level
            self.levels = level.levels
        if isinstance(level, Tuple):
            self.factor = level[0]
            if isinstance(level[1], List):
                self.levels = level[1]
            else:
                self.levels = [level[1]]
        for i in range(len(self.levels)):
            if isinstance(self.levels[i], Level):
                self.levels[i] = self.levels[i].name


class AtMostKInARow(_KConstraint):

    def test(self, sequence: List):
        name = self.factor.name
        for level in self.levels:
            atMost = 0
            for trial in sequence:
                if trial[name] == level:
                    atMost += 1
                    if atMost > self.trials:
                        return False
                else:
                    atMost = 0
        return True


class AtLeastKInARow(_KConstraint):

    def test(self, sequence: List):
        name = self.factor.name
        for level in self.levels:
            atLeast = 0
            for trial in sequence:
                if trial[name] == level:
                    atLeast += 1
                else:
                    if atLeast < self.trials and atLeast != 0:
                        return False
                    atLeast = 0
        return True


class ExactlyKInARow(_KConstraint):

    def test(self, sequence: List):
        is_exact = True
        name = self.factor.name
        for level in self.levels:
            indexes = []
            for i in range(len(sequence)):
                trial = sequence[i]
                if trial[name] == level and (i < 0 or sequence[i - 1][name] != level):
                    indexes.append(i)
                if trial[name] == level and (i >= len(sequence) - 1 or sequence[i + 1][name] != level):
                    indexes.append(i)
            for i in range(1, len(indexes), 2):
                is_exact = is_exact and indexes[i] - indexes[i - 1] == self.trials - 1
        return is_exact


class ExactlyK(_KConstraint):

    def test(self, sequence: List):
        name = self.factor.name
        for level in self.levels:
            nr = 0
            for trial in sequence:
                if trial[name] == level:
                    nr += 1
            if nr != self.trials:
                return False
        return True


class Block:
    design: List[Factor]
    crossing: List[Factor]
    constraints: List[Constraint]
    _counter_balanced_levels: List[Level]
    _counter_balanced_names_weights: List

    def __init__(self, design: List[Factor] = None, crossing: List[Factor] = None,
                 constraints: List[Constraint] = None):
        self.design = design
        self.crossing = crossing
        self.constraints = constraints
        if not self.design:
            self.design = []
        if not self.crossing:
            self.crossing = []
        if not self.constraints:
            self.constraints = []
        # get exclude constraints:
        exclude_constraints = {}
        for factor in self.design:
            exclude_constraints[factor.name] = []
        for c in self.constraints:
            if isinstance(c, Exclude):
                exclude_constraints[c.factor] += [c.level]

        if self.crossing:
            levels = [[lvl] if not lvl.name in exclude_constraints[self.crossing[0].name] else [] for lvl in
                      self.crossing[0].levels]
            i = 1
            while i < len(self.crossing):
                list_2 = self.crossing[i].levels
                tmp = [x + [y] if not y.name in exclude_constraints[self.crossing[i].name] else [] for x in levels
                       for y in list_2]
                levels = tmp
                i += 1
            self._counter_balanced_levels = levels
            # self._counter_balanced_levels = list(filter(lambda x: x['name'] != [], self._counter_balanced_levels))
            self._counter_balanced_names_weights = []
            for level in levels:
                name = [lvl.name for lvl in level]
                weight = 1
                for lvl in level:
                    weight *= lvl.weight
                res = {'name': name, 'weight': weight}
                self._counter_balanced_names_weights.append(res)
        self._counter_balanced_levels = list(filter(lambda x: x != [], self._counter_balanced_levels))
        self._counter_balanced_names_weights = list(
            filter(lambda x: x['name'] != [], self._counter_balanced_names_weights))

    def test(self, sequence: List):
        chi_2 = self._test_crossing(sequence)
        derived_test = self._test_levels(sequence)
        constraint_test = []
        for c in self.constraints:
            constraint_test.append(c.test(sequence))
        return {'pValue': chi_2.pvalue, 'levels': derived_test, 'constraints': constraint_test}

    def _test_crossing(self, sequence: List):
        if not self.crossing:
            return chisquare([1, 1], [1, 1])
        weights_empirical = [0 for _ in self._counter_balanced_names_weights]
        for t in sequence:
            level = []
            for factor in self.crossing:
                level.append(t[factor.name])
            i = 0
            for lvl in self._counter_balanced_names_weights:
                if level == lvl['name']:
                    weights_empirical[i] += 1
                else:
                    i += 1
        weights_expected = [lvl['weight'] for lvl in self._counter_balanced_names_weights]

        # normalize weights
        # adjusting for the first trials if there is a transition window in crossing
        max_width = 0
        for factor in self.crossing:
            for lvl in factor.levels:
                if isinstance(lvl, DerivedLevel):
                    if lvl.window.width > 1:
                        max_width = max(lvl.window.width - 1, max_width)

        total_expected = sum(weights_expected) + max_width
        weights_expected = [w / total_expected * len(sequence) for w in weights_expected]

        sum_empirical = sum(weights_empirical)
        sum_expected = sum(weights_expected)
        weights_expected = [sum_empirical * w / sum_expected for w in weights_expected]
        return chisquare(weights_empirical, weights_expected)

    def _test_levels(self, sequence: List):
        test = {}
        for factor in self.design:
            test[factor.name] = True
            is_derived = False
            for lvl in factor.levels:
                if isinstance(lvl, DerivedLevel):
                    is_derived = True
            if is_derived:
                for i in range(len(sequence)):
                    trial = sequence[i]
                    for lvl in factor.levels:
                        if isinstance(lvl, DerivedLevel):
                            window = lvl.window
                            if window.width == 1:
                                args = [trial[factor_w.name] for factor_w in window.factors]
                                lvl_t = trial[factor.name]
                                if window.predicate(*args):
                                    test[factor.name] = (lvl_t == lvl.name) and test[factor.name]
                            elif i >= (window.width - 1):
                                s = i - (window.width - 1)
                                e = i + 1
                                sequence_ = sequence[s:e]
                                args = [[trial[factor_w.name] for trial in sequence_] for factor_w in window.factors]
                                lvl_t = sequence_[-1][factor.name]
                                if window.predicate(*args):
                                    test[factor.name] = (lvl_t == lvl.name) and test[factor.name]
                                    if (lvl_t != lvl.name):
                                        print(i, sequence_)
        return test



from sourpea.primitives import Level, Factor, Block

# counterbalance 1
# sequence to test on
test_sequence_1 = [
    {'word': 'red', 'color': 'red'},
    {'word': 'green', 'color': 'red'},
    {'word': 'red', 'color': 'green'},
    {'word': 'green', 'color': 'green'},
]

# defining the factors
word = Factor('word', ['red', 'green'])
color = Factor('color', ['red', 'green'])

# defining the design (counterbalancing)
block_1 = Block(design=[word, color], crossing=[word, color])

# getting the results
test_1 = block_1.test(test_sequence_1)
print('sequence1, design1: ', test_1['pValue'], test_1['levels'])

# counterbalance with weighted factors
# sequence to test on
test_sequence_2 = [
    {'word': 'red', 'color': 'red'},
    {'word': 'green', 'color': 'red'},
    {'word': 'red', 'color': 'red'},
    {'word': 'green', 'color': 'green'},
    {'word': 'red', 'color': 'green'},
    {'word': 'red', 'color': 'green'}
]

# defining the factors
word_2 = Factor('word', [Level('red', 2), 'green'])
color_2 = Factor('color', ['red', 'green'])

# defining the design
block_2 = Block(design=[color_2, word_2], crossing=[word_2, color_2])

# getting the result
test_2 = block_2.test(test_sequence_2)
print('sequence2, design2: ', test_2['pValue'], test_2['levels'])

# counterbalance with more weights
# sequence to test on
test_sequence_3 = [
    {'word': 'red', 'color': 'green'},
    {'word': 'red', 'color': 'green'},
    {'word': 'red', 'color': 'green'},
    {'word': 'red', 'color': 'green'},
    {'word': 'red', 'color': 'green'},
    {'word': 'red', 'color': 'green'},
    {'word': 'red', 'color': 'red'},
    {'word': 'red', 'color': 'red'},
    {'word': 'red', 'color': 'red'},
    {'word': 'red', 'color': 'red'},
    {'word': 'green', 'color': 'green'},
    {'word': 'green', 'color': 'green'},
    {'word': 'green', 'color': 'green'},
    {'word': 'green', 'color': 'red'},
    {'word': 'green', 'color': 'red'},
]

# defining factors
word_3 = Factor('word', [Level('red', 2), Level('green', 1)])
color_3 = Factor('color', [Level('red', 2), Level('green', 3)])

# defining design
block_3 = Block(design=[word_3, color_3], crossing=[word_3, color_3])

# getting the results
test_3 = block_3.test(test_sequence_3)
print('sequence3, design3: ', test_3['pValue'], test_3['levels'])

# cross validating (this should result in lower p-values)
test_4 = block_1.test(test_sequence_2)
print('sequence2, design1: ', test_4['pValue'], test_4['levels'])

test_5 = block_2.test(test_sequence_1)
print('sequence1, design2: ', test_5['pValue'], test_5['levels'])



from sourpea.primitives import Factor, Level, Block, MinimumTrials, AtMostKInARow, AtLeastKInARow, ExactlyKInARow, \
    ExactlyK

test_sequence = [
    {'direction': 'up', 'orientation': 'down', 'congruency': 'incongruent'},
    {'direction': 'down', 'orientation': 'down', 'congruency': 'congruent'},
    {'direction': 'up', 'orientation': 'up', 'congruency': 'incongruent'},
    {'direction': 'down', 'orientation': 'up', 'congruency': 'congruent'},
]

test_sequence_2 = [
    {'direction': 'up', 'orientation': 'down', 'congruency': 'incongruent'},
    {'direction': 'up', 'orientation': 'up', 'congruency': 'congruent'},
    {'direction': 'up', 'orientation': 'up', 'congruency': 'incongruent'},
    {'direction': 'down', 'orientation': 'up', 'congruency': 'congruent'},
]

direction = Factor(name='direction', initial_levels=['up', 'down'])
orientation = Factor(name='orientation', initial_levels=['up', 'down'])

minimum_trials = MinimumTrials(4)
minimum_trials_2 = MinimumTrials(8)
atMostKInARow = AtMostKInARow(2, (direction, ['up']))
atMostKInARow_2 = AtMostKInARow(2, direction)
atLeastKInARow = AtLeastKInARow(2, orientation)
atLeastKInARow_2 = AtLeastKInARow(2, (orientation, 'up'))
exactlyKInARow = ExactlyKInARow(2, orientation)
exactlyKInARow_2 = ExactlyKInARow(3, (direction, Level('up')))
exactlyK = ExactlyK(3, (direction, 'up'))

block = Block(constraints=[minimum_trials, atMostKInARow, atLeastKInARow, exactlyKInARow, exactlyK])
block_2 = Block(constraints=[minimum_trials_2, atMostKInARow_2, atLeastKInARow_2, exactlyKInARow_2, exactlyK])

test = block.test(test_sequence)
test_2 = block.test(test_sequence_2)
test_3 = block_2.test(test_sequence)
test_4 = block_2.test(test_sequence_2)

print(test)
print(test_2)
print(test_3)
print(test_4)



from sourpea.primitives import DerivedLevel, Level, Factor, Block, WithinTrialDerivationWindow, \
    TransitionDerivationWindow, Exclude

# valid test sequence
test_sequence_1 = [
    {'word': 'red', 'color': 'red', 'congruency': 'con'},
    {'word': 'green', 'color': 'green', 'congruency': 'con'},
]

# invalid test sequence: derived level mixup
test_sequence_2 = [
    {'word': 'red', 'color': 'red', 'congruency': 'inc'},
    {'word': 'red', 'color': 'green', 'congruency': 'inc'},
    {'word': 'green', 'color': 'red', 'congruency': 'con'},
    {'word': 'green', 'color': 'green', 'congruency': 'con'},
]

# valid test
test_sequence_3 = [
    {'word': 'red', 'color': 'red', 'congruency': 'con'},
    {'word': 'red', 'color': 'red', 'congruency': 'con'},
    {'word': 'green', 'color': 'green', 'congruency': 'con'},
    {'word': 'green', 'color': 'green', 'congruency': 'con'},
]

word = Factor('word', ['red', 'green'])
color = Factor('color', ['red', 'green'])


def is_con(w, c):
    return w == c


def is_inc(w, c):
    return not is_con(w, c)


congruent = DerivedLevel('con', window=WithinTrialDerivationWindow(predicate=is_con, factors=[word, color]))
incongruent = DerivedLevel('inc', window=WithinTrialDerivationWindow(predicate=is_inc, factors=[word, color]))

congruency = Factor('congruency', [congruent, incongruent])

exclude = Exclude('congruency', 'inc')

block = Block(design=[word, color, congruency], crossing=[word, congruency], constraints=[exclude])
test_1 = block.test(test_sequence_1)
test_2 = block.test(test_sequence_2)
test_3 = block.test(test_sequence_3)

print(test_1['pValue'], test_1['levels'])
print(test_2['pValue'], test_2['levels'])
print(test_3['pValue'], test_3['levels'])
